package es.ucm.fdi.grasia.faerie.core;

import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.attributeInstanceURI;

import java.net.URI;

import org.apache.log4j.Logger;

/**
 * Default implementation of the {@code Attribute<E, T>} interface. It provides proper synchronization and notifies the
 * changes on the {@code ContextValue}.
 *
 * @author Chema
 * @param <E> the type of the entity.
 * @param <T> the type of the value.
 */
public abstract class AbstractAttribute<E extends Entity, T> extends AbstractContextElement implements Attribute<E, T> {

	private static final long serialVersionUID = 8508180717708123636L;

	private E entity;

	private final Boolean valueLock = false;
	private ContextValue<T> value = null;

	private boolean remotelyUpdateable = false;

	public AbstractAttribute() {
		entity = null;
	}

	/**
	 * Construct an attribute instance associated with the specified {@code entity}.
	 *
	 * @param entity the associated {@code Entity}.
	 */
	public AbstractAttribute(E entity) {
		this.setId(attributeInstanceURI(this.getClass(), entity.getId()));
		this.setEntity(entity);
	}

	/**
	 * Construct an attribute instance with the specified {@code id} and associated with the specified {@code entity}.
	 *
	 * @param id the URI of the attribute instance.
	 * @param entity the associated {@code Entity}
	 */
	public AbstractAttribute(URI id, E entity) {
		super(id);
		this.setEntity(entity);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public E getEntity() {
		return entity;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void setEntity(E entity) {
		if (entity != null)
			if (this.entity == null || entity.getClass().equals(this.entity.getClass())
					&& entity.getId().equals(this.entity.getId())) {
				this.entity = entity;
				if (this.getId() == null)
					this.setId(attributeInstanceURI(this.getClass(), entity.getId()));
			} else
				throw new IllegalArgumentException("entity " + entity + " not compatible with " + this);
		else
			throw new NullPointerException("entity is null");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ContextValue<T> getValue() {
		synchronized (valueLock) {
			return value;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ContextValue<T> getValueWaitIfNull() throws InterruptedException {
		synchronized (valueLock) {
			while (value == null)
				valueLock.wait();
			return value;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ContextValue<T> getValueWaitIfNull(long timeout) throws InterruptedException {
		synchronized (valueLock) {
			if (value == null)
				valueLock.wait(timeout);
			return value;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ContextValue<T> update(ContextValue<T> value) {
		synchronized (valueLock) {
			ContextValue<T> oldValue = this.value;
			if (oldValue == null || !oldValue.equalsWithoutDate(value)) {
				this.value = value;
				Logger.getLogger(this.getClass()).debug("Updated * " + this);
				valueLock.notifyAll();
				this.notifyElementUpdated(ContextEvent.VALUE_UPDATED, oldValue);
			}
			return oldValue;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isRemotelyUpdateable() {
		return remotelyUpdateable;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setRemotelyUpdateable(boolean remote) {
		this.remotelyUpdateable = remote;
	}

	/**
	 * Returns the string representation of this attribute instance. It consist on the URi followed with the
	 * {@code ContextValue}.
	 *
	 * @return the string representation of this attribute instance.
	 */
	@Override
	public String toString() {
		synchronized (valueLock) {
			return entity + "#" + this.getClass().getSimpleName() + " [" + (value != null ? value.getData() : null)
					+ "]";
		}
	}
}
