package es.ucm.fdi.grasia.faerie.core;

import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.felix.scr.annotations.Component;
import org.apache.log4j.Logger;

/**
 * Default implementation for an {@code AbstractContextObserverService} that also updates certain context elements.
 *
 * @author Chema
 * @see Context
 * @see Attribute
 */
@Component(inherit = true, componentAbstract = true)
@SuppressWarnings({ "unchecked", "rawtypes" })
public abstract class AbstractAttributeUpdaterComponent extends AbstractContextObserverComponent implements
		ContextUpdaterDefinition {

	private final Class<? extends Attribute>[] updatedContextElementTypes = setUpUpdatedContextElementTypes();

	private final Map<URI, Attribute> updatedAttributeInstances = setUpUpdatedAttributeInstances();
	private final Map<URI, ContextElement> linkedElements = setUpLinkedElements();
	private final Map<URI, Set<URI>> elementsLinkedAttributeInstances = setUpElementsLinkedAttributeInstances();
	private final Map<URI, Set<URI>> attributeInstancesLinkedElements = setUpAttributeInstancesLinkedElements();

	private final LinkedElementObserver linkedElementsObserver = setUpLinkedElementsObserver();

	protected Class<? extends Attribute>[] setUpUpdatedContextElementTypes() {
		return new Class[0];
	}

	protected Map<URI, Attribute> setUpUpdatedAttributeInstances() {
		return new HashMap<URI, Attribute>();
	}

	protected Map<URI, ContextElement> setUpLinkedElements() {
		return new HashMap<URI, ContextElement>();
	}

	protected Map<URI, Set<URI>> setUpElementsLinkedAttributeInstances() {
		return new HashMap<URI, Set<URI>>();
	}

	protected HashMap<URI, Set<URI>> setUpAttributeInstancesLinkedElements() {
		return new HashMap<URI, Set<URI>>();
	}

	protected LinkedElementObserver setUpLinkedElementsObserver() {
		return new LinkedElementObserver();
	}

	@Override
	public Class<? extends ContextElement>[] getUpdatedContextElementTypes() {
		return updatedContextElementTypes;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void deactivate() {
		removeAllUpdatedAttributeInstances();
		super.deactivate();
	}

	@Override
	public void elementRemoved(Attribute attributeInstance, ContextEvent event) {
		removeUpdatedAttributeInstance(attributeInstance);
	}

	/**
	 * Returns an attribute instance being updated.
	 *
	 * @param elementId the id of the attribute instance.
	 * @return the attribute instance being updated with the specified id.
	 */
	protected <A extends Attribute> A getUpdatedAttributeInstance(URI elementId) {
		synchronized (updatedAttributeInstances) {
			return (A) updatedAttributeInstances.get(elementId);
		}
	}

	/**
	 * Returns all attribute instances being updated.
	 *
	 * @return all attribute instances being updated.
	 */
	protected Map<URI, Attribute> getAllUpdatedAttributeInstances() {
		synchronized (updatedAttributeInstances) {
			Map<URI, Attribute> clone = setUpUpdatedAttributeInstances();
			clone.putAll(updatedAttributeInstances);
			return clone;
		}
	}

	protected Map<URI, Attribute> searchAttributeInstancesByEntityId(URI entityId) {
		synchronized (updatedAttributeInstances) {
			Map<URI, Attribute> clone = setUpUpdatedAttributeInstances();
			for (Entry<URI, Attribute> entry : updatedAttributeInstances.entrySet())
				if (entry.getValue().getEntity().getId().equals(entityId))
					clone.put(entry.getKey(), entry.getValue());
			return clone;
		}
	}

	/**
	 * Sets an attribute instance to be updated.
	 *
	 * @param attributeInstance the attribute instance to be updated.
	 */
	protected void putUpdatedAttributeInstance(Attribute attributeInstance) {
		synchronized (updatedAttributeInstances) {
			if (updatedAttributeInstances.put(attributeInstance.getId(), attributeInstance) == null)
				Logger.getLogger(this.getClass()).debug(
						"Put Updating: " + attributeInstance + " * " + this.strHashCode());
		}
	}

	/**
	 * Removes an attribute instance from being updated.
	 *
	 * @param elementId the URI of the attribute instance.
	 */
	protected void removeUpdatedAttributeInstance(URI elementId) {
		synchronized (updatedAttributeInstances) {
			Attribute removed;
			if ((removed = updatedAttributeInstances.remove(elementId)) != null)
				Logger.getLogger(this.getClass()).debug(
						"Quit Updating: " + removed + " * " + this.strHashCode());
			Set<URI> attributeInstanceLinkedElementsIds = attributeInstancesLinkedElements.remove(elementId);
			if (attributeInstanceLinkedElementsIds != null)
				for (URI attributeInstanceLinkedElementId : attributeInstanceLinkedElementsIds) {
					Set<URI> elementLinkedAttributeInstances = elementsLinkedAttributeInstances
							.get(attributeInstanceLinkedElementId);
					elementLinkedAttributeInstances.remove(elementId);
					if (elementLinkedAttributeInstances.isEmpty()) {
						elementsLinkedAttributeInstances.remove(attributeInstanceLinkedElementId);
						linkedElements.remove(attributeInstanceLinkedElementId).removeObserver(linkedElementsObserver);
						this.contextRelease(attributeInstanceLinkedElementId);
					}
				}
		}
	}

	/**
	 * Removes an attribute instance from being updated.
	 *
	 * @param attributeInstance the attribute instance to be removed.
	 */
	protected void removeUpdatedAttributeInstance(Attribute attributeInstance) {
		removeUpdatedAttributeInstance(attributeInstance.getId());
	}

	/**
	 * Removes all attribute instances from being updated.
	 */
	protected void removeAllUpdatedAttributeInstances() {
		synchronized (updatedAttributeInstances) {
			for (URI attributeInstanceId : getAllUpdatedAttributeInstances().keySet())
				removeUpdatedAttributeInstance(attributeInstanceId);
		}
	}

	/**
	 * Returns all the elements which changes produces changes to the specified attribute instance.
	 *
	 * @param attributeInstanceId the URI of the attribute instance.
	 * @return all the elements which changes produces changes to the specified attribute instance.
	 */
	protected Map<URI, ContextElement> getAttributeInstanceLinkedElements(URI attributeInstanceId) {
		synchronized (updatedAttributeInstances) {
			Set<URI> linkedElementsIds = attributeInstancesLinkedElements.get(attributeInstanceId);
			Map<URI, ContextElement> result = setUpLinkedElements();
			if (linkedElementsIds != null)
				for (URI linkedElementId : linkedElementsIds)
					result.put(linkedElementId, linkedElements.get(linkedElementId));
			return result;
		}
	}

	Set<URI> setUpIdSet() {
		return new HashSet<URI>();
	}

	/**
	 * Links an element to an updated attribute instance, so that the changes of the linked element produces the update
	 * of the updated attribute instance.
	 *
	 * @param attributeInstanceId the URI of the attibute instance to be updated.
	 * @param linkedElement the linked element.
	 * @return the previous linked element with the same URI.
	 */
	protected ContextElement addAttributeInstanceLinkedElement(URI attributeInstanceId, ContextElement linkedElement) {
		synchronized (updatedAttributeInstances) {
			URI linkedElementId = linkedElement.getId();
			Set<URI> attributeInstancesIds = elementsLinkedAttributeInstances.get(linkedElementId);
			if (attributeInstancesIds == null) {
				linkedElement = this.contextRequest(linkedElement);
				linkedElements.put(linkedElementId, linkedElement);
				elementsLinkedAttributeInstances.put(linkedElementId, attributeInstancesIds = setUpIdSet());
				linkedElement.addObserver(linkedElementsObserver);
			} else
				linkedElement = this.contextPeek(linkedElementId);
			attributeInstancesIds.add(attributeInstanceId);

			Set<URI> linkedElementsIds = attributeInstancesLinkedElements.get(attributeInstanceId);
			if (linkedElementsIds == null)
				attributeInstancesLinkedElements.put(attributeInstanceId, linkedElementsIds = setUpIdSet());
			if (linkedElementsIds.add(linkedElementId))
				Logger.getLogger(this.getClass()).debug(
						"Linking: " + updatedAttributeInstances.get(attributeInstanceId) + " <- " + linkedElement
								+ " * " + this.strHashCode());

			return linkedElement;
		}
	}

	protected void removeAttributeInstanceLinkedElement(URI attributeInstanceId, ContextElement linkedElement) {
		synchronized (updatedAttributeInstances) {
			URI linkedElementId = linkedElement.getId();

			Set<URI> linkedElementsIds = attributeInstancesLinkedElements.get(attributeInstanceId);
			if (linkedElementsIds != null)
				if (linkedElementsIds.remove(linkedElementId))
					Logger.getLogger(this.getClass()).debug(
							"Unlinking: " + updatedAttributeInstances.get(attributeInstanceId) + "<-/-" + linkedElement
									+ " * " + this.strHashCode());

			Set<URI> attributeInstancesIds = elementsLinkedAttributeInstances.get(linkedElementId);
			if (attributeInstancesIds != null) {
				attributeInstancesIds.remove(attributeInstanceId);
				if (attributeInstancesIds.isEmpty()) {
					elementsLinkedAttributeInstances.remove(linkedElementId);
					linkedElements.remove(linkedElementId).removeObserver(linkedElementsObserver);
					this.contextRelease(linkedElementId);
				}
			}
		}
	}

	/**
	 * Updates the attribute instance with the specified id.
	 *
	 * @param attributeInstanceId the URI of the attribute instance.
	 */
	protected void updateAttributeInstance(URI attributeInstanceId) {
		synchronized (updatedAttributeInstances) {
			Attribute updatedAttributeInstance = updatedAttributeInstances.get(attributeInstanceId);
			attributeInstanceLinkedElementsUpdated(updatedAttributeInstance,
					getAttributeInstanceLinkedElements(attributeInstanceId));
		}
	}

	/**
	 * Updates all the attribute instances linked with the specified element.
	 *
	 * @param elementId the URI of the element.
	 */
	protected void updateLinkedAttributeInstances(URI elementId) {
		synchronized (updatedAttributeInstances) {
			Set<URI> attributeInstanceIds = elementsLinkedAttributeInstances.get(elementId);
			Logger.getLogger(this.getClass()).debug(
					"Updating: " + attributeInstanceIds + " <- " + elementId + " * " + this.strHashCode());
			if (attributeInstanceIds != null)
				for (URI attributeInstanceId : attributeInstanceIds)
					updateAttributeInstance(attributeInstanceId);
		}
	}

	/**
	 * Called when any linked element of a certain attribute instance is modified.
	 *
	 * @param attributeInstance the attribute instance.
	 * @param linkedElements all the linked elements.
	 */
	protected void attributeInstanceLinkedElementsUpdated(Attribute attributeInstance,
			Map<URI, ContextElement> linkedElements) {
		// Empty
	}

	protected class LinkedElementObserver extends ContextEventAdapter {

		@Override
		public void elementUpdated(ContextEvent event) {
			updateLinkedAttributeInstances(event.getSource().getId());
		}
	}
}
