package es.ucm.fdi.grasia.faerie.core;

import java.net.URI;
import java.util.HashSet;
import java.util.Set;

import org.apache.felix.scr.annotations.Activate;
import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Deactivate;
import org.apache.felix.scr.annotations.Reference;
import org.apache.log4j.Logger;

import es.ucm.fdi.grasia.faerie.core.model.BundleEntity;
import es.ucm.fdi.grasia.faerie.core.model.EnvironmentReferencedEntitiesSet;
import es.ucm.fdi.grasia.faerie.core.model.IncludedEntities;

/**
 * Default implementation for a service running in the environment. It provides access to the {@code ContextManager}
 * wich in turn gives access to the {@code Context} and other framework utils.
 *
 * @author Chema
 * @see Context
 * @see Entity
 */
@Component(componentAbstract = true)
public abstract class AbstractContextComponent implements ContextComponent {

	private final Boolean contextBrokerLock = false;
	@Reference
	private ContextBroker contextBroker = null;

	private IncludedEntities<Entity> environmentReferencedEntities = null;

	private final Set<Entity> myReferencedEntities = setUpMyReferencedEntities();

	protected Set<Entity> setUpMyReferencedEntities() {
		return new HashSet<Entity>();
	}

	protected String strHashCode() {
		return Integer.toHexString(this.hashCode());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void bindContextBroker(ContextBroker contextBroker) {
		synchronized (contextBrokerLock) {
			Logger.getLogger(this.getClass()).trace("(" + contextBroker + ") * " + strHashCode());
			this.contextBroker = contextBroker;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Activate
	@Override
	public void activate() {
		String classifier = (this instanceof ContextObserver) ? ContextObserver.class.getSimpleName()
				: ContextComponent.class.getSimpleName();
		Logger.getLogger(this.getClass()).debug("(" + classifier + ") Activating * " + strHashCode());

		requestEnvironmentReferencedEntities();

		addMyReferencedEntities();

		doActivate();
	}

	private void requestEnvironmentReferencedEntities() {
		environmentReferencedEntities = contextRequest(new IncludedEntities<Entity>(
				EnvironmentReferencedEntitiesSet.getInstance()));
	}

	protected void addMyReferencedEntities() {
	}

	protected void doActivate() {
	}

	/**
	 * {@inheritDoc}
	 */
	@Deactivate
	@Override
	public void deactivate() {
		String classifier = (this instanceof ContextObserver) ? ContextObserver.class.getSimpleName()
				: ContextComponent.class.getSimpleName();
		Logger.getLogger(this.getClass()).debug("(" + classifier + ") Deactivating * " + strHashCode());

		removeMyReferencedEntities();

		releaseEnvironmentReferencedEntities();
	}

	protected void removeMyReferencedEntities() {
		synchronized (myReferencedEntities) {
			for (Entity entity : getAllMyReferencedEntities())
				removeReferencedEntity(entity);
		}
	}

	private void releaseEnvironmentReferencedEntities() {
		contextRelease(environmentReferencedEntities.getId());
		environmentReferencedEntities = null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void unbindContextBroker(ContextBroker contextManager) {
		synchronized (contextBrokerLock) {
			Logger.getLogger(this.getClass()).trace("(" + contextManager + ") * " + strHashCode());
			this.contextBroker = null;
		}
	}

	/**
	 * Returns the context manager of this service.
	 *
	 * @return the context manager of this service.
	 */
	protected ContextBroker getContextBroker() {
		synchronized (contextBrokerLock) {
			return contextBroker;
		}
	}

	/**
	 * {@link ContextManager#getContext()}
	 */
	protected Context getContext() {
		synchronized (contextBrokerLock) {
			if (contextBroker != null)
				return contextBroker.getContext();
			throw new NotAccessibleContextException("contextManager is null");
		}
	}

	/**
	 * {@link ContextManager#getServiceEntity()}
	 */
	public BundleEntity getBundleEntity() {
		synchronized (contextBrokerLock) {
			if (contextBroker != null)
				return contextBroker.getBundleEntity();
			throw new NotAccessibleContextException("contextManager is null");
		}
	}

	/**
	 * {@link Context#contains(java.net.URI)}
	 */
	protected boolean contextContains(URI elementId) {
		synchronized (contextBrokerLock) {
			return getContext().contains(elementId);
		}
	}

	/**
	 * {@link Context#peek(java.net.URI)}
	 */
	protected <T extends ContextElement> T contextPeek(URI elementId) {
		synchronized (contextBrokerLock) {
			return getContext().peek(elementId);
		}
	}

	/**
	 * {@link Context#request(net.faerie.core.ContextElement)}
	 */
	protected <T extends ContextElement> T contextRequest(T element) {
		synchronized (contextBrokerLock) {
			String classifier = (element instanceof Entity) ? Entity.class.getSimpleName()
					: (element instanceof Attribute) ? Attribute.class.getSimpleName() : Relationship.class
							.getSimpleName();
			Logger.getLogger(this.getClass())
					.debug("Requesting " + classifier + ": " + element + " * " + strHashCode());
			return getContext().request(element);
		}
	}

	/**
	 * {@link Context#release(java.net.URI)}
	 */
	protected boolean contextRelease(URI elementId) {
		synchronized (contextBrokerLock) {
			Logger.getLogger(this.getClass()).debug("Releasing id: " + elementId + " * " + strHashCode());
			return getContext().release(elementId);
		}
	}

	/**
	 * {@link Context#releaseAll(java.net.URI)}
	 */
	protected void contextReleaseAll(URI elementId) {
		synchronized (contextBrokerLock) {
			Logger.getLogger(this.getClass()).debug("Releasing All id: " + elementId + " * " + strHashCode());
			getContext().releaseAll(elementId);
		}
	}

	/**
	 * {@link Context#releaseAll()}
	 */
	protected void contextReleaseAll() {
		synchronized (contextBrokerLock) {
			Logger.getLogger(this.getClass()).debug("Releasing All * " + strHashCode());
			getContext().releaseAll();
		}
	}

	protected IncludedEntities<Entity> getEnvironmentReferencedEntities() {
		return environmentReferencedEntities;
	}

	protected Set<Entity> getAllMyReferencedEntities() {
		synchronized (myReferencedEntities) {
			Set<Entity> clone = setUpMyReferencedEntities();
			clone.addAll(myReferencedEntities);
			return clone;
		}
	}

	protected <E extends Entity> E addReferencedEntity(E entity) {
		synchronized (myReferencedEntities) {
			if (myReferencedEntities.add(entity)) {
				entity = contextRequest(entity);
				environmentReferencedEntities.addTargetEntity(entity);
			}
			return entity;
		}
	}

	protected void removeReferencedEntity(Entity entity) {
		synchronized (myReferencedEntities) {
			if (myReferencedEntities.remove(entity)) {
				environmentReferencedEntities.removeTargetEntity(entity);
				contextRelease(entity.getId());
			}
		}
	}

}
