package es.ucm.fdi.grasia.faerie.core;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.URI;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * ! hecho
 *
 * Default implementation of {@code ContextElement} interface. It offers protected methods for notification of
 * {@code ContextEvent}s to the {@code ContextObserverCallback}s.
 *
 * @author Chema
 * @see ContextEvent
 */
public abstract class AbstractContextElement implements ContextElement {

	private static final long serialVersionUID = 199014031004200887L;
	private URI id;
	private final Boolean observersLock = false;
	transient private Set<ContextObserverCallback> observers = setUpObservers();
	transient private Context context;

	protected Set<ContextObserverCallback> setUpObservers() {
		return new LinkedHashSet<ContextObserverCallback>();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		observers = setUpObservers();
	}

	AbstractContextElement() {
		id = null;
		context = null;
	}

	/**
	 * Constructs an element with the specified id.
	 *
	 * @param id the id representing this element.
	 */
	AbstractContextElement(URI id) {
		this.setId(id);
		context = null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public URI getId() {
		return id;
	}

	protected final void setId(URI id) {
		if (id == null)
			throw new NullPointerException("id is null");
		this.id = id;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean addObserver(ContextObserverCallback observer) {
		synchronized (observersLock) {
			return observers.add(observer);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean removeObserver(ContextObserverCallback observer) {
		synchronized (observersLock) {
			return observers.remove(observer);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int countObservers() {
		synchronized (observersLock) {
			return observers.size();
		}
	}

	/**
	 * Returns a copy of the set of observers of this element.
	 *
	 * @return a copy of the set of observers of this element.
	 */
	protected Set<ContextObserverCallback> getAllObservers() {
		synchronized (observersLock) {
			Set<ContextObserverCallback> clone = setUpObservers();
			clone.addAll(observers);
			return clone;
		}
	}

	/**
	 * Clears all the {@code ContextObserver}s of this element.
	 */
	protected void removeAllObservers() {
		synchronized (observersLock) {
			observers.clear();
		}
	}

	protected ContextEvent createContextEvent(int eventType, Object arg) {
		return new ContextEvent(this, eventType, arg);
	}

	protected class NotifyElementUpdatedTask implements Runnable {

		private ContextObserverCallback observer;
		private int eventType;
		private Object arg;

		protected NotifyElementUpdatedTask(ContextObserverCallback observer, int eventType, Object arg) {
			this.observer = observer;
			this.eventType = eventType;
			this.arg = arg;
		}

		@Override
		public void run() {
			ContextEvent event = createContextEvent(eventType, arg);
			if ((observer.handle(event) & ContextObserver.HANDLE_READ) != 0)
				observer.elementUpdated(event);
		}
	}

	protected Runnable createNotifyElementUpdatedTask(ContextObserverCallback observer, int eventType, Object arg) {
		return new NotifyElementUpdatedTask(observer, eventType, arg);
	}

	/**
	 * Notifies a {@code ContextEvent} from this element with the specified event type and argument.
	 *
	 * @param eventType the event type of the {@code ContextEvent}.
	 * @param arg the argument to attach to the {@code ContextEvent}.
	 * @see ContextEvent
	 */
	protected void notifyElementUpdated(int eventType, Object arg) {
		synchronized (observersLock) {
			for (ContextObserverCallback observer : observers)
				FaerieExecutorService.getExecutorService().execute(
						createNotifyElementUpdatedTask(observer, eventType, arg));
		}
	}

	@Override
	public Context getContext() {
		return context;
	}

	protected final void setContext(Context context) {
		this.context = context;
	}
}
