package es.ucm.fdi.grasia.faerie.core;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Service;
import org.apache.log4j.Logger;

/**
 * Default implementation for a {@code AbstractContextService} that is a {@code ContextObserver} in the system. This
 * class is to be extended by developers.
 *
 * @author Chema
 * @see Context
 * @see ContextElement
 */
@Service(value = ContextObserver.class)
@Component(inherit = true, componentAbstract = true)
@SuppressWarnings({ "unchecked", "rawtypes" })
public abstract class AbstractContextObserverComponent extends AbstractContextComponent implements ContextObserver {

	private final Class<? extends ContextElement>[] observedContextElementTypes = setUpObservedContextElementTypes();
	private final Map<URI, ContextElement> observedElements = setUpObservedElements();
	private final Map<Object, Behavior> behaviors = setUpBehaviors();

	protected Class<? extends ContextElement>[] setUpObservedContextElementTypes() {
		return new Class[] {};
	}

	protected Map<URI, ContextElement> setUpObservedElements() {
		return new HashMap<URI, ContextElement>();
	}

	protected Map<Object, Behavior> setUpBehaviors() {
		return new HashMap<Object, Behavior>();
	}

	@Override
	public Class<? extends ContextElement>[] getObservedContextElementTypes() {
		return observedContextElementTypes;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void deactivate() {
		removeAllObservedElements();
		super.deactivate();
	}

	/**
	 * Returns the observed element with the specified URI.
	 *
	 * @param <T> the type of the element.
	 * @param elementId the URI of the element.
	 * @return the observed element with the specified URI.
	 */
	protected <T extends ContextElement> T getObservedElement(URI elementId) {
		synchronized (observedElements) {
			return (T) observedElements.get(elementId);
		}
	}

	/**
	 * Returns a copy of all observed elements, mapped by URI.
	 *
	 * @return a copy of all observed elements, mapped by URI.
	 */
	protected Map<URI, ContextElement> getAllObservedElements() {
		synchronized (observedElements) {
			Map<URI, ContextElement> clone = setUpObservedElements();
			clone.putAll(observedElements);
			return clone;
		}
	}

	/**
	 * Puts an element to be observed.
	 *
	 * @param observedElement the element to be observed.
	 */
	protected void putObservedElement(ContextElement observedElement) {
		synchronized (observedElements) {
			if (observedElements.put(observedElement.getId(), observedElement) == null)
				Logger.getLogger(this.getClass()).debug(
						"Put Observing: " + observedElement + " * " + this.strHashCode());
			observedElement.addObserver(this);
		}
	}

	/**
	 * Remove an element from being observed.
	 *
	 * @param elementId the URI of the element.
	 */
	protected void removeObservedElement(URI elementId) {
		synchronized (observedElements) {
			ContextElement element = observedElements.get(elementId);
			if (element != null)
				removeObservedElement(element);
		}
	}

	/**
	 * Remove an element from being observed.
	 *
	 * @param observedElement the element to be removed.
	 */
	protected void removeObservedElement(ContextElement observedElement) {
		synchronized (observedElements) {
			observedElement.removeObserver(this);
			if (observedElements.remove(observedElement.getId()) != null)
				Logger.getLogger(this.getClass()).debug(
						"Quit Observing: " + observedElement + " * " + this.strHashCode());
		}
	}

	/**
	 * Removes all elements from being observed.
	 */
	protected void removeAllObservedElements() {
		synchronized (observedElements) {
			for (ContextElement contextElement : getAllObservedElements().values())
				removeObservedElement(contextElement);
		}
	}

	protected Behavior getBehavior(ContextEvent event) {
		Attribute attributeInstance = event.getSourceAttributeInstance();
		ContextValue value = attributeInstance.getValue();
		Object data = (value != null) ? value.getData() : null;
		return behaviors.get(data);
	}

	protected void putBehavior(Object key, Behavior behavior) {
		synchronized (behaviors) {
			behaviors.put(key, behavior);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int handle(ContextEvent event) {
		// Default
		return (event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) == 0
				? ContextObserver.HANDLE_READ : ContextObserver.HANDLE_NONE;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void elementAdded(ContextEvent event) {
		Logger.getLogger(this.getClass()).debug("Received: " + event + " * " + this.strHashCode());

		ContextElement element = event.getContextElementArg();
		if (element instanceof Entity)
			elementAdded((Entity) element, event);
		else if (element instanceof Attribute)
			elementAdded((Attribute) element, event);
		else if (element instanceof Relationship)
			elementAdded((Relationship) element, event);
	}

	/**
	 * Called when the observer is notified of a new {@code Entity} in the context.
	 *
	 * @param entity the new {@code Entity}.
	 * @param event the associated event.
	 */
	public void elementAdded(Entity entity, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a new {@code Attribute} in the context.
	 *
	 * @param attributeInstance the new {@code Attribute}
	 * @param event the associated event.
	 */
	public void elementAdded(Attribute attributeInstance, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a new {@code Relationship} in the context.
	 *
	 * @param relationshipInstance the new {@code Relationship}.
	 * @param event the associated event.
	 */
	public void elementAdded(Relationship relationshipInstance, ContextEvent event) {
		// Empty
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void elementRemoved(ContextEvent event) {
		Logger.getLogger(this.getClass()).debug("Received: " + event + " * " + this.strHashCode());

		ContextElement element = event.getContextElementArg();
		if (element instanceof Entity)
			elementRemoved((Entity) element, event);
		else if (element instanceof Attribute)
			elementRemoved((Attribute) element, event);
		else if (element instanceof Relationship)
			elementRemoved((Relationship) element, event);
	}

	/**
	 * Called when the observer is notified of a removed {@code Entity}.
	 *
	 * @param entity the removed Entity.
	 * @param event the associated event.
	 */
	public void elementRemoved(Entity entity, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a removed {@code Attribute}.
	 *
	 * @param attributeInstance the removed Attribute.
	 * @param event the associated event.
	 */
	public void elementRemoved(Attribute attributeInstance, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a removed {@code Relationship}.
	 *
	 * @param relationshipInstance the removed Relationship.
	 * @param event the associated event.
	 */
	public void elementRemoved(Relationship relationshipInstance, ContextEvent event) {
		// Empty
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void elementUpdated(ContextEvent event) {
		Logger.getLogger(this.getClass()).debug("Received: " + event + " * " + this.strHashCode());

		ContextElement element = event.getSource();
		if (element instanceof Entity)
			elementUpdated((Entity) element, event);
		else if (element instanceof Attribute)
			elementUpdated((Attribute) element, event);
		else if (element instanceof Relationship)
			elementUpdated((Relationship) element, event);
	}

	/**
	 * Called when the observer is notified of a updated {@code Entity}.
	 *
	 * @param entity the updated Entity.
	 * @param event the associated event.
	 */
	public void elementUpdated(Entity entity, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a updated {@code Attribute}.
	 *
	 * @param attributeInstance the updated Attribute.
	 * @param event the associated event.
	 */
	public void elementUpdated(Attribute attributeInstance, ContextEvent event) {
		// Empty
	}

	/**
	 * Called when the observer is notified of a updated {@code Relationship}.
	 *
	 * @param relationshipInstance the updated Relationship.
	 * @param event the associated event.
	 */
	public void elementUpdated(Relationship relationshipInstance, ContextEvent event) {
		// Empty
	}
}
