package es.ucm.fdi.grasia.faerie.core;

import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.authorityURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.entityURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.localEntityURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.nameURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.namespaceURI;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import es.ucm.fdi.grasia.faerie.core.util.ContextURIs;

/**
 * ! hecho
 * 
 * Default implementation for the {@code Entity} interface. It holds collections for the associated
 * {@code Attribute(Instance)}s and {@code Relationship(Instance)}s.
 * 
 * @author Chema
 * @see Attribute
 * @see Relationship
 */
@SuppressWarnings({ "unchecked", "rawtypes" })
public abstract class AbstractEntity extends AbstractContextElement implements Entity {

	private static final long serialVersionUID = -7756068177755209992L;

	private final Boolean attributesInstancesLock = false;
	transient private Map<Class<? extends Attribute>, Attribute> attributesInstances = setUpAttributesInstances();

	private final Boolean relationshipsInstancesLock = false;
	transient private Map<Class<? extends Relationship>, Relationship> relationshipsInstances = setUpRelationshipsInstances();

	transient private AttributesInstancesObserver attsInstancesObserver = setUpAttsInstancesObserver();
	transient private RelationshipsInstancesObserver relsInstancesObserver = setUpRelsInstancesObserver();

	protected Map<Class<? extends Attribute>, Attribute> setUpAttributesInstances() {
		return new HashMap<Class<? extends Attribute>, Attribute>();
	}

	protected Map<Class<? extends Relationship>, Relationship> setUpRelationshipsInstances() {
		return new HashMap<Class<? extends Relationship>, Relationship>();
	}

	protected AttributesInstancesObserver setUpAttsInstancesObserver() {
		return new AttributesInstancesObserver();
	}

	protected RelationshipsInstancesObserver setUpRelsInstancesObserver() {
		return new RelationshipsInstancesObserver();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		attributesInstances = setUpAttributesInstances();
		relationshipsInstances = setUpRelationshipsInstances();
		attsInstancesObserver = setUpAttsInstancesObserver();
		relsInstancesObserver = setUpRelsInstancesObserver();
	}

	public AbstractEntity() {
	}

	public AbstractEntity(URI namespace, URI name) {
		super(localEntityURI(namespace, name));
	}

	/**
	 * Constructs an entity of the specified authority, namespace and name, and the default type.
	 * 
	 * @param authority the container that created the entity.
	 * @param namespace the namespace of the application.
	 * @param name the name of the entity.
	 */
	public AbstractEntity(URI authority, URI namespace, URI name) {
		super(entityURI(authority, namespace, name));
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public URI getAuthority() {
		URI id = this.getId();
		if (id != null)
			return authorityURI(id);
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public URI getNamespace() {
		URI id = this.getId();
		if (id != null)
			return namespaceURI(id);
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public URI getName() {
		URI id = this.getId();
		if (id != null)
			return nameURI(id);
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public URI getLocalId() {
		URI result = this.getId();
		if (result != null)
			result = URI.create("/").resolve(ContextURIs.localURI().relativize(result));
		return result;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <A extends Attribute> A getAttributeInstance(Class<A> attribute) {
		synchronized (attributesInstancesLock) {
			return (A) attributesInstances.get(attribute);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Map<Class<? extends Attribute>, Attribute> getAllAttributesInstances() {
		synchronized (attributesInstancesLock) {
			Map<Class<? extends Attribute>, Attribute> clone = setUpAttributesInstances();
			clone.putAll(attributesInstances);
			return clone;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <A extends Attribute> A putAttributeInstance(A attributeInstance) {
		synchronized (attributesInstancesLock) {
			if (attributeInstance.getEntity() != this)
				throw new IllegalArgumentException("attributeInstance " + attributeInstance + " not compatible with "
						+ this);
			A oldAttInstance = (A) attributesInstances.put(attributeInstance.getClass(), attributeInstance);
			if (attributeInstance != oldAttInstance) {
				attributeInstance.addObserver(attsInstancesObserver);
				this.notifyElementUpdated(ContextEvent.ATTINSTANCE_PUT, attributeInstance);
			}
			return oldAttInstance;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <A extends Attribute> A removeAttributeInstance(Class<A> attribute) {
		synchronized (attributesInstancesLock) {
			A ai = (A) attributesInstances.remove(attribute);
			if (ai != null) {
				ai.removeObserver(attsInstancesObserver);
				this.notifyElementUpdated(ContextEvent.ATTINSTANCE_REMOVED, ai);
			}
			return ai;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void removeAllAttributesInstances() {
		synchronized (attributesInstancesLock) {
			if (!attributesInstances.isEmpty()) {
				Map<Class<? extends Attribute>, Attribute> oldAttInstances = getAllAttributesInstances();
				for (Attribute ai : oldAttInstances.values())
					ai.removeObserver(attsInstancesObserver);
				attributesInstances.clear();
				this.notifyElementUpdated(ContextEvent.ALL_ATTINSTANCES_REMOVED, oldAttInstances);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <R extends Relationship> R getRelationshipInstance(Class<R> relationship) {
		synchronized (relationshipsInstancesLock) {
			return (R) relationshipsInstances.get(relationship);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Map<Class<? extends Relationship>, Relationship> getAllRelationshipsInstances() {
		synchronized (relationshipsInstancesLock) {
			Map<Class<? extends Relationship>, Relationship> clone = setUpRelationshipsInstances();
			clone.putAll(relationshipsInstances);
			return clone;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <R extends Relationship> R putRelationshipInstance(R relationshipInstance) {
		synchronized (relationshipsInstancesLock) {
			if (relationshipInstance.getSourceEntity() != this)
				throw new IllegalArgumentException("relationshipInstance " + relationshipInstance
						+ " not compatible with " + this);
			R oldRelInstance = (R) relationshipsInstances.put(relationshipInstance.getClass(), relationshipInstance);
			if (relationshipInstance != oldRelInstance) {
				relationshipInstance.addObserver(relsInstancesObserver);
				this.notifyElementUpdated(ContextEvent.RELINSTANCE_PUT, relationshipInstance);
			}
			return oldRelInstance;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <R extends Relationship> R removeRelationshipInstance(Class<R> relationship) {
		synchronized (relationshipsInstancesLock) {
			R ri = (R) relationshipsInstances.remove(relationship);
			if (ri != null) {
				ri.removeObserver(relsInstancesObserver);
				this.notifyElementUpdated(ContextEvent.RELINSTANCE_REMOVED, ri);
			}
			return ri;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void removeAllRelationshipsInstances() {
		synchronized (relationshipsInstancesLock) {
			if (!relationshipsInstances.isEmpty()) {
				Map<Class<? extends Relationship>, Relationship> oldRelInstances = getAllRelationshipsInstances();
				for (Relationship ri : oldRelInstances.values())
					ri.removeObserver(relsInstancesObserver);
				relationshipsInstances.clear();
				this.notifyElementUpdated(ContextEvent.ALL_RELINSTANCES_REMOVED, oldRelInstances);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		return getLocalId() + ":" + this.getClass().getSimpleName();
	}

	// These classes redirect the notifications from the associated elements to this element's observers.
	protected class AttributesInstancesObserver extends ContextEventAdapter implements Serializable {

		private static final long serialVersionUID = 4526176052469152203L;

		@Override
		public void elementUpdated(ContextEvent event) {
			AbstractEntity.this.notifyElementUpdated(ContextEvent.ATTINSTANCE_UPDATED, event);
		}
	}

	protected class RelationshipsInstancesObserver extends ContextEventAdapter implements Serializable {

		private static final long serialVersionUID = 1990528620245820650L;

		@Override
		public void elementUpdated(ContextEvent event) {
			AbstractEntity.this.notifyElementUpdated(ContextEvent.RELINSTANCE_UPDATED, event);
		}
	}
}
