package es.ucm.fdi.grasia.faerie.core;

import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.felix.scr.annotations.Component;
import org.apache.log4j.Logger;

/**
 * Default implementation for an {@code AbstractContextObserverService} that also updates certain context elements.
 *
 * @author Chema
 * @see Context
 * @see Relationship
 */
@Component(inherit = true, componentAbstract = true)
@SuppressWarnings({ "unchecked", "rawtypes" })
public abstract class AbstractRelationshipUpdaterComponent extends AbstractContextObserverComponent implements
		ContextUpdaterDefinition {

	private final Class<? extends Relationship>[] updatedContextElementTypes = setUpUpdatedContextElementTypes();

	private final Map<URI, Relationship> updatedRelationshipInstances = setUpUpdatedRelationshipInstances();
	private final Map<URI, ContextElement> linkedElements = setUpLinkedElements();
	private final Map<URI, Set<URI>> elementsLinkedRelationshipInstances = setUpElementsLinkedRelationshipInstances();
	private final Map<URI, Set<URI>> relationshipInstancesLinkedElements = setUpRelationshipInstancesLinkedElements();

	private final LinkedElementObserver linkedElementsObserver = setUpLinkedElementsObserver();

	protected Class<? extends Relationship>[] setUpUpdatedContextElementTypes() {
		return new Class[0];
	}

	protected Map<URI, Relationship> setUpUpdatedRelationshipInstances() {
		return new HashMap<URI, Relationship>();
	}

	protected Map<URI, ContextElement> setUpLinkedElements() {
		return new HashMap<URI, ContextElement>();
	}

	protected Map<URI, Set<URI>> setUpElementsLinkedRelationshipInstances() {
		return new HashMap<URI, Set<URI>>();
	}

	protected HashMap<URI, Set<URI>> setUpRelationshipInstancesLinkedElements() {
		return new HashMap<URI, Set<URI>>();
	}

	protected LinkedElementObserver setUpLinkedElementsObserver() {
		return new LinkedElementObserver();
	}

	@Override
	public Class<? extends ContextElement>[] getUpdatedContextElementTypes() {
		return updatedContextElementTypes;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void deactivate() {
		removeAllUpdatedRelationshipInstances();
		super.deactivate();
	}

	@Override
	public void elementRemoved(Relationship relationshipInstance, ContextEvent event) {
		removeUpdatedRelationshipInstance(relationshipInstance);
	}

	/**
	 * Returns an relationship instance being updated.
	 *
	 * @param elementId the id of the relationship instance.
	 * @return the relationship instance being updated with the specified id.
	 */
	protected <R extends Relationship> R getUpdatedRelationshipInstance(URI elementId) {
		synchronized (updatedRelationshipInstances) {
			return (R) updatedRelationshipInstances.get(elementId);
		}
	}

	/**
	 * Returns all relationship instances being updated.
	 *
	 * @return all relationship instances being updated.
	 */
	protected Map<URI, Relationship> getAllUpdatedRelationshipInstances() {
		synchronized (updatedRelationshipInstances) {
			Map<URI, Relationship> clone = setUpUpdatedRelationshipInstances();
			clone.putAll(updatedRelationshipInstances);
			return clone;
		}
	}

	protected Map<URI, Relationship> searchRelationshipInstancesBySourceEntityId(URI entityId) {
		synchronized (updatedRelationshipInstances) {
			Map<URI, Relationship> clone = setUpUpdatedRelationshipInstances();
			for (Entry<URI, Relationship> entry : updatedRelationshipInstances.entrySet())
				if (entry.getValue().getSourceEntity().getId().equals(entityId))
					clone.put(entry.getKey(), entry.getValue());
			return clone;
		}
	}

	/**
	 * Sets an relationship instance to be updated.
	 *
	 * @param relationshipInstance the relationship instance to be updated.
	 */
	protected void putUpdatedRelationshipInstance(Relationship relationshipInstance) {
		synchronized (updatedRelationshipInstances) {
			if (updatedRelationshipInstances.put(relationshipInstance.getId(), relationshipInstance) == null)
				Logger.getLogger(this.getClass()).debug(
						"Put Updating: " + relationshipInstance + " * " + this.strHashCode());
		}
	}

	/**
	 * Removes an relationship instance from being updated.
	 *
	 * @param elementId the URI of the relationship instance.
	 */
	protected void removeUpdatedRelationshipInstance(URI elementId) {
		synchronized (updatedRelationshipInstances) {
			if (updatedRelationshipInstances.remove(elementId) != null)
				Logger.getLogger(this.getClass()).debug(
						"Quit Updating: " + updatedRelationshipInstances.get(elementId) + " * " + this.strHashCode());
			Set<URI> relationshipInstanceLinkedElementsIds = relationshipInstancesLinkedElements.remove(elementId);
			if (relationshipInstanceLinkedElementsIds != null)
				for (URI relationshipInstanceLinkedElementId : relationshipInstanceLinkedElementsIds) {
					Set<URI> elementLinkedRelationshipInstances = elementsLinkedRelationshipInstances
							.get(relationshipInstanceLinkedElementId);
					elementLinkedRelationshipInstances.remove(elementId);
					if (elementLinkedRelationshipInstances.isEmpty()) {
						elementsLinkedRelationshipInstances.remove(relationshipInstanceLinkedElementId);
						linkedElements.remove(relationshipInstanceLinkedElementId).removeObserver(
								linkedElementsObserver);
						this.contextRelease(relationshipInstanceLinkedElementId);
					}
				}
		}
	}

	/**
	 * Removes an relationship instance from being updated.
	 *
	 * @param relationshipInstance the relationship instance to be removed.
	 */
	protected void removeUpdatedRelationshipInstance(Relationship<?, ?> relationshipInstance) {
		removeUpdatedRelationshipInstance(relationshipInstance.getId());
	}

	/**
	 * Removes all relationship instances from being updated.
	 */
	protected void removeAllUpdatedRelationshipInstances() {
		synchronized (updatedRelationshipInstances) {
			for (URI relationshipInstanceId : getAllUpdatedRelationshipInstances().keySet())
				removeUpdatedRelationshipInstance(relationshipInstanceId);
		}
	}

	/**
	 * Returns all the elements which changes produces changes to the specified relationship instance.
	 *
	 * @param relationshipInstanceId the URI of the relationship instance.
	 * @return all the elements which changes produces changes to the specified relationship instance.
	 */
	protected Map<URI, ContextElement> getRelationshipInstanceLinkedElements(URI relationshipInstanceId) {
		synchronized (updatedRelationshipInstances) {
			Set<URI> linkedElementsIds = relationshipInstancesLinkedElements.get(relationshipInstanceId);
			Map<URI, ContextElement> result = setUpLinkedElements();
			if (linkedElementsIds != null)
				for (URI linkedElementId : linkedElementsIds)
					result.put(linkedElementId, linkedElements.get(linkedElementId));
			return result;
		}
	}

	Set<URI> setUpIdSet() {
		return new HashSet<URI>();
	}

	/**
	 * Links an element to an updated relationship instance, so that the changes of the linked element produces the
	 * update of the updated relationship instance.
	 *
	 * @param relationshipInstanceId the URI of the relationship instance to be updated.
	 * @param linkedElement the linked element.
	 * @return the previous linked element with the same URI.
	 */
	protected ContextElement addRelationshipInstanceLinkedElement(URI relationshipInstanceId,
			ContextElement linkedElement) {
		synchronized (updatedRelationshipInstances) {
			URI linkedElementId = linkedElement.getId();
			Set<URI> relationshipInstancesIds = elementsLinkedRelationshipInstances.get(linkedElementId);
			if (relationshipInstancesIds == null) {
				linkedElement = this.contextRequest(linkedElement);
				linkedElements.put(linkedElementId, linkedElement);
				elementsLinkedRelationshipInstances.put(linkedElementId, relationshipInstancesIds = setUpIdSet());
				linkedElement.addObserver(linkedElementsObserver);
			} else
				linkedElement = this.contextPeek(linkedElementId);
			relationshipInstancesIds.add(relationshipInstanceId);

			Set<URI> linkedElementsIds = relationshipInstancesLinkedElements.get(relationshipInstanceId);
			if (linkedElementsIds == null)
				relationshipInstancesLinkedElements.put(relationshipInstanceId, linkedElementsIds = setUpIdSet());
			if (linkedElementsIds.add(linkedElementId))
				Logger.getLogger(this.getClass()).debug(
						"Linking: " + updatedRelationshipInstances.get(relationshipInstanceId) + " <- " + linkedElement
								+ " * " + this.strHashCode());

			return linkedElement;
		}
	}

	protected void removeRelationshipInstanceLinkedElement(URI attributeInstanceId, ContextElement linkedElement) {
		synchronized (updatedRelationshipInstances) {
			URI linkedElementId = linkedElement.getId();

			Set<URI> linkedElementsIds = relationshipInstancesLinkedElements.get(attributeInstanceId);
			if (linkedElementsIds != null)
				if (linkedElementsIds.remove(linkedElementId))
					Logger.getLogger(this.getClass()).debug(
							"Unlinking: " + updatedRelationshipInstances.get(attributeInstanceId) + "<-/-"
									+ linkedElement + " * " + this.strHashCode());

			Set<URI> attributeInstancesIds = elementsLinkedRelationshipInstances.get(linkedElementId);
			if (attributeInstancesIds != null) {
				attributeInstancesIds.remove(attributeInstanceId);
				if (attributeInstancesIds.isEmpty()) {
					elementsLinkedRelationshipInstances.remove(linkedElementId);
					linkedElements.remove(linkedElementId).removeObserver(linkedElementsObserver);
					this.contextRelease(linkedElementId);
				}
			}
		}
	}

	/**
	 * Updates the relationship instance with the specified id.
	 *
	 * @param relationshipInstanceId the URI of the relationship instance.
	 */
	protected void updateRelationshipInstance(URI relationshipInstanceId) {
		synchronized (updatedRelationshipInstances) {
			Relationship updatedRelationshipInstance = updatedRelationshipInstances.get(relationshipInstanceId);
			if (updatedRelationshipInstance != null)
				relationshipInstanceLinkedElementsUpdated(updatedRelationshipInstance,
						getRelationshipInstanceLinkedElements(relationshipInstanceId));
		}
	}

	/**
	 * Updates all the relationship instances linked with the specified element.
	 *
	 * @param elementId the URI of the element.
	 */
	protected void updateLinkedRelationshipInstances(URI elementId) {
		synchronized (updatedRelationshipInstances) {
			Set<URI> relationshipInstanceIds = elementsLinkedRelationshipInstances.get(elementId);
			Logger.getLogger(this.getClass()).debug(
					"Updating: " + relationshipInstanceIds + " <- " + elementId + " * " + this.strHashCode());
			if (relationshipInstanceIds != null)
				for (URI relationshipInstanceId : relationshipInstanceIds)
					updateRelationshipInstance(relationshipInstanceId);
		}
	}

	/**
	 * Called when any linked element of a certain relationship instance is modified.
	 *
	 * @param relationshipInstance the relationship instance.
	 * @param linkedElements all the linked elements.
	 */
	protected void relationshipInstanceLinkedElementsUpdated(Relationship relationshipInstance,
			Map<URI, ContextElement> linkedElements) {
		// Empty
	}

	protected class LinkedElementObserver extends ContextEventAdapter {

		@Override
		public void elementUpdated(ContextEvent event) {
			updateLinkedRelationshipInstances(event.getSource().getId());
		}
	}
}
