package es.ucm.fdi.grasia.faerie.core;

/**
 * Represents an instance of an Attribute applied to an {@code Entity} with certain value.
 * 
 * @author Chema
 * @param <E> the type of entity to which the attribute is applied.
 * @param <T> the type of value of the attribute.
 * @see Entity
 * @see ContextValue
 */
public interface Attribute<E extends Entity, V> extends ContextElement {

	/**
	 * Returns the entity for which this instance is applied.
	 * 
	 * @return the entity for which this instance is applied.
	 */
	public E getEntity();

	/**
	 * Not to be used by applications. Sets the entity for which this instance is applied.
	 * 
	 * @param entity the entity for which this instance is applied.
	 */
	public void setEntity(E entity);

	/**
	 * Returns the value of the Attribute applied to the {@code Entity}.
	 * 
	 * @return the value of the Attribute applied to the {@code Entity}.
	 */
	public ContextValue<V> getValue();

	/**
	 * Returns the value of the Attribute applied to the {@code Entity}. The calling thread is blocked if the value is
	 * null.
	 * 
	 * @return the value of the Attribute applied to the {@code Entity}.
	 * @throws InterruptedException if the thread was interrupted while waiting.
	 */
	public ContextValue<V> getValueWaitIfNull() throws InterruptedException;

	/**
	 * Returns the value of the Attribute applied to the {@code Entity}. The calling thread is blocked if the value is
	 * null a maximum of timeout milliseconds.
	 * 
	 * @param timeout the maximum waiting time in milliseconds.
	 * @return the value of the Attribute applied to the {@code Entity}.
	 * @throws InterruptedException if the thread was interrupted while waiting.
	 */
	public ContextValue<V> getValueWaitIfNull(long timeout) throws InterruptedException;

	/**
	 * Updates the {@code ContextValue} of this {@code AttributeInstance} and returns the previous value. The element
	 * notifies its observers if the specified value is not equals to the previous value.
	 * 
	 * @param value the new value of the attribute.
	 * @return the previous value of the attribute.
	 */
	public ContextValue<V> update(ContextValue<V> value);

	/**
	 * Returns true if the attribute instance should be calculated remotely, false otherwise.
	 * 
	 * @return true if the attribute instance should be calculated remotely, false otherwise.
	 */
	public boolean isRemotelyUpdateable();

	/**
	 * Sets if this attribute instance should be remotely calculated. The default is true.
	 * 
	 * @param remotelyUpdateable true if the attribute instance should be calculated remotely, false otherwise.
	 */
	public void setRemotelyUpdateable(boolean remotelyUpdateable);
}
