package es.ucm.fdi.grasia.faerie.core;

import java.net.URI;

/**
 * Represents the context of a certain environment. It manages the access to de context elements from a single component
 * of the system, restricting the {@link ContextContainer} interface as if the client was the only within the system.
 * 
 * <h4>{@code Contextobserver}s and request/release mechanism</h4>
 * 
 * When the client invokes the {@link #request(net.faerie.core.ContextElement)} method, the context does the following:
 * <ol>
 * <li>It search an existing element with the same {@code elementId}.</li>
 * <li>If the element exists:
 * <ol>
 * <li>Returns the existing element.</li>
 * </ol>
 * otherwise:
 * <ol>
 * <li>It sets the request for the element to zero.</li>
 * <li>It adds the element to the context.</li>
 * <li>It calls the {@link net.faerie.core.ContextObserver#elementAdded(net.faerie.core.ContextEvent)} method of all
 * registered {@code ContextObserver}s, provided that the method
 * {@link net.faerie.core.ContextObserver#handles(net.faerie.core.ContextEvent)} for the element returns true.</li>
 * </ol>
 * </li>
 * <li>It increases the request count for the element.</li>
 * </ol>
 * 
 * When a client invokes the {@link #release(java.net.URI)} method, the context container does the following:
 * <ol>
 * <li>It decreases the request count for the element.</li>
 * <li>If the request count has reached zero:
 * <ol>
 * <li>It calls the {@link net.faerie.core.ContextObserver#elementRemoved(net.faerie.core.ContextEvent)} method of all
 * registered {@code ContextObserver}s, provided that the method
 * {@link net.faerie.core.ContextObserver#handles(net.faerie.core.ContextEvent)} for the event returns true.</li>
 * <li>It removes the element from the context.</li>
 * </ol>
 * </li>
 * </ol>
 * 
 * @author Chema
 * @see ContextElement
 * @see ContextManager
 * @see ContextObserver
 */
public interface Context {

	/**
	 * Test if the context contains the element with the specified {@code elementId}.
	 * 
	 * @param elementId the URI of the element.
	 * @return true if the element is contained in the context, false otherwise.
	 */
	public boolean contains(URI elementId);

	/**
	 * Obtains the element with the specified {@code elementId} if it exists in the context. Returns {@code null}
	 * otherwise.
	 * 
	 * @param <T> the subclass of the element.
	 * @param elementId the URI of the element.
	 * @return the element with the specified URI, if exists in the context, or null otherwise.
	 */
	public <T extends ContextElement> T peek(URI elementId);

	/**
	 * Requests the specified {@code element}.
	 * 
	 * @param <T> the subclass of the element.
	 * @param element the element requested.
	 * @return the specified element, following the rules described at the beginning.
	 */
	public <T extends ContextElement> T request(T element);

	/**
	 * Releases the {@code ContextElement} with the specified {@code elementId}.
	 * 
	 * @param elementId the URI of the element.
	 * @return true if the request count has reached zero, false otherwise.
	 */
	public boolean release(URI elementId);

	/**
	 * Puts the request count of the specified element to zero for the calling component.
	 * 
	 * @param elementId the URI of the element.
	 */
	public void releaseAll(URI elementId);

	/**
	 * Releases all the {@code ContextElement}s requested by the client.
	 */
	public void releaseAll();
}
