package es.ucm.fdi.grasia.faerie.core;

import java.io.Serializable;

/**
 * Represents an event produced in the context. It has a type and an associated information.
 * 
 * @author Chema
 * @see ContextElement
 * @see ContextObserver
 * @see Attribute
 * @see ContextValue
 * @see Relationship
 * @see Entity
 */
@SuppressWarnings({ "rawtypes" })
public class ContextEvent implements Serializable {

	private static final long serialVersionUID = 8071652546445169490L;

	private static final int UNKNOWN_EVENT = 0;
	private static final String UNKNOWN_EVENT_STR = "UNKNOWN_EVENT";
	/**
	 * An element has been added to the context. The argument is that element.
	 */
	public static final int ELEMENT_ADDED = 1;
	public static final String ELEMENT_ADDED_STR = "ELEMENT_ADDED";
	/**
	 * An element has been removed from the context. The argument is that element.
	 */
	public static final int ELEMENT_REMOVED = 2;
	public static final String ELEMENT_REMOVED_STR = "ELEMENT_REMOVED";
	/**
	 * An {@code AttributeInstance} has been put to the source {@code Entity}. The argument is that
	 * {@code AttributeInstance}.
	 */
	public static final int ATTINSTANCE_PUT = 4;
	public static final String ATTINSTANCE_PUT_STR = "ATTINSTANCE_PUT";
	/**
	 * An {@code AttributeInstance} from the source {@code Entity} has been updated. The argument is the
	 * {@code ContextEvent} produced by the {@code AttributeInstance}.
	 */
	public static final int ATTINSTANCE_UPDATED = 8;
	public static final String ATTINSTANCE_UPDATED_STR = "ATTINSTANCE_UPDATED";
	/**
	 * An {@code AttributeInstance} from the source {@code Entity} has been removed. The argument is the removed
	 * {@code AttributeInstance}.
	 */
	public static final int ATTINSTANCE_REMOVED = 16;
	public static final String ATTINSTANCE_REMOVED_STR = "ATTINSTANCE_REMOVED";
	/**
	 * All {@code AttributeInstance}s from the source {@code Entity} has been removed. The argument is the collection of
	 * removed {@code AttributeInstance}s.
	 */
	public static final int ALL_ATTINSTANCES_REMOVED = 32;
	public static final String ALL_ATTINSTANCES_REMOVED_STR = "ALL_ATTINSTANCES_REMOVED";
	/**
	 * A {@code RelationshipInstance} has been put to the source {@code Entity}. The argument is that
	 * {@code RelationshipInstance}.
	 */
	public static final int RELINSTANCE_PUT = 64;
	public static final String RELINSTANCE_PUT_STR = "RELINSTANCE_PUT";
	/**
	 * A {@code RelationshipInstance} from the source {@code Entity} has been updated. The argument is the
	 * {@code ContextEvent} produced by the {@code AttributeInstance}.
	 */
	public static final int RELINSTANCE_UPDATED = 128;
	public static final String RELINSTANCE_UPDATED_STR = "RELINSTANCE_UPDATED";
	/**
	 * A {@code RelationshipInstance} from the source {@code Entity} has been removed. The argument is the removed
	 * {@code RelationshipInstance}.
	 */
	public static final int RELINSTANCE_REMOVED = 256;
	public static final String RELINSTANCE_REMOVED_STR = "RELINSTANCE_REMOVED";
	/**
	 * All {@code RelationshipInstance}s from the source {@code Entity} has been removed. The argument is the collection
	 * of removed {@code RelationshipInstance}s.
	 */
	public static final int ALL_RELINSTANCES_REMOVED = 512;
	public static final String ALL_RELINSTANCES_REMOVED_STR = "ALL_RELINSTANCES_REMOVED";
	/**
	 * The {@code ContextValue} of the source {@code AttributeInstance} has been modified. The argument is the old
	 * {@code ContextValue}
	 */
	public static final int VALUE_UPDATED = 1024;
	public static final String VALUE_UPDATED_STR = "VALUE_UPDATED";
	/**
	 * An {@code Entity} has been put as target to the source {@code RelationshipInstance}. The argument is that
	 * {@code Entity}.
	 */
	public static final int TARGETENTITY_PUT = 2048;
	public static final String TARGETENTITY_PUT_STR = "TARGETENTITY_PUT";
	/**
	 * All the target entities has been reset.
	 */
	public static final int ALL_TARGETENTITIES_PUT = 4096;
	public static final String ALL_TARGETENTITIES_PUT_STR = "ALL_TARGETENTITIES_PUT";
	/**
	 * An {@code Entity} has been removed as target from the source {@code RelationshipInstance}. The argument is that
	 * {@code Entity}.
	 */
	public static final int TARGETENTITY_REMOVED = 8192;
	public static final String TARGETENTITY_REMOVED_STR = "TARGETENTITY_REMOVED";
	/**
	 * All target {@code Entity}s has been removed from the source {@code RelationshipInstance}. The argument is the
	 * collection of removed {@code Entity}s.
	 */
	public static final int ALL_TARGETENTITIES_REMOVED = 16384;
	public static final String ALL_TARGETENTITIES_REMOVED_STR = "ALL_TARGETENTITIES_REMOVED";

	private final ContextElement source;
	private final int eventType;
	private final Object arg;

	/**
	 * Constructs an event specifying a source, an event type and an argument.
	 * 
	 * @param source the element that produced the event.
	 * @param eventType the type of the event.
	 * @param arg an argument associated to the event.
	 */
	public ContextEvent(ContextElement source, int eventType, Object arg) {
		this.source = source;
		this.eventType = eventType;
		this.arg = arg;
	}

	/**
	 * Returns the source that produced the event.
	 * 
	 * @return the source that produced the event.
	 */
	public ContextElement getSource() {
		return source;
	}

	/**
	 * Tries to cast the source {@code ContextElement} to an {@code Entity}. Returns null if the source is not an
	 * instance of {@code Entity}.
	 * 
	 * @return the source element if it is an instance of {@code Entity}, or null otherwise.
	 */
	public Entity getSourceEntity() {
		try {
			return (Entity) source;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Tries to cast the source {@code ContextElement} to an {@code AttributeInstance}. Returns null if the source is
	 * not an instance of {@code AttributeInstance}.
	 * 
	 * @return the source element if it is an instance of {@code AttributeInstance}, or null otherwise.
	 */
	public Attribute getSourceAttributeInstance() {
		try {
			return (Attribute) source;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Tries to cast the source {@code ContextElement} to an {@code RelationshipInstance}. Returns null if the source is
	 * not an instance of {@code RelationshipInstance}.
	 * 
	 * @return the source element if it is an instance of {@code RelationshipInstance}, or null otherwise.
	 */
	public Relationship getSourceRelationshipInstance() {
		try {
			return (Relationship) source;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Returns the type code of the event.
	 * 
	 * @return the type code of the event.
	 */
	public int getEventType() {
		return eventType;
	}

	/**
	 * Returns the associated argument of the event.
	 * 
	 * @return the associated argument of the event.
	 */
	public Object getArg() {
		return arg;
	}

	/**
	 * Tries to cast the argument to a {@code ContextElement}. Returns null if it is not an instance of
	 * {@code ContextElement}.
	 * 
	 * @return the argument if it is an instance of {@code ContextElement}, or null otherwise.
	 */
	public ContextElement getContextElementArg() {
		try {
			return (ContextElement) arg;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Tries to cast the argument to a {@code Entity}. Returns null if it is not an instance of {@code Entity}.
	 * 
	 * @return the argument if it is an instance of {@code Entity}, or null otherwise.
	 */
	public Entity getEntityArg() {
		try {
			return (Entity) arg;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Tries to cast the argument to a {@code AttributeInstance}. Returns null if it is not an instance of
	 * {@code AttributeInstance}.
	 * 
	 * @return the argument if it is an instance of {@code AttributeInstance}, or null otherwise.
	 */
	public Attribute getAttributeInstanceArg() {
		try {
			return (Attribute) arg;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * Tries to cast the argument to a {@code RelationshipInstance}. Returns null if it is not an instance of
	 * {@code RelationshipInstance}.
	 * 
	 * @return the argument if it is an instance of {@code RelationshipInstance}, or null otherwise.
	 */
	public Relationship getRelationshipInstanceArg() {
		try {
			return (Relationship) arg;
		} catch (ClassCastException ex) {
			return null;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int hashCode() {
		int hash = 5;
		hash = 59 * hash + (this.source != null ? this.source.hashCode() : 0);
		hash = 59 * hash + this.eventType;
		hash = 59 * hash + (this.arg != null ? this.arg.hashCode() : 0);
		return hash;
	}

	/**
	 * Returns true if this event represents the same event as the specified object.
	 * 
	 * @param obj the object to be compared.
	 * @return true if and only if the specified object represents the same event.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		if (this.getClass() != obj.getClass())
			return false;
		ContextEvent other = (ContextEvent) obj;
		if (this.source != other.source)
			return false;
		if (this.eventType != other.eventType)
			return false;
		if (this.arg != other.arg && (this.arg == null || !this.arg.equals(other.arg)))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return source + " -> " + toString(eventType) + "(" + arg + ")";
	}

	public static String toString(int eventType) {
		switch (eventType) {
		case ELEMENT_ADDED:
			return ELEMENT_ADDED_STR;
		case ELEMENT_REMOVED:
			return ELEMENT_REMOVED_STR;
		case ATTINSTANCE_PUT:
			return ATTINSTANCE_PUT_STR;
		case ATTINSTANCE_UPDATED:
			return ATTINSTANCE_UPDATED_STR;
		case ATTINSTANCE_REMOVED:
			return ATTINSTANCE_REMOVED_STR;
		case ALL_ATTINSTANCES_REMOVED:
			return ALL_ATTINSTANCES_REMOVED_STR;
		case RELINSTANCE_PUT:
			return RELINSTANCE_PUT_STR;
		case RELINSTANCE_UPDATED:
			return RELINSTANCE_UPDATED_STR;
		case RELINSTANCE_REMOVED:
			return RELINSTANCE_REMOVED_STR;
		case ALL_RELINSTANCES_REMOVED:
			return ALL_RELINSTANCES_REMOVED_STR;
		case VALUE_UPDATED:
			return VALUE_UPDATED_STR;
		case TARGETENTITY_PUT:
			return TARGETENTITY_PUT_STR;
		case ALL_TARGETENTITIES_PUT:
			return ALL_TARGETENTITIES_PUT_STR;
		case TARGETENTITY_REMOVED:
			return TARGETENTITY_REMOVED_STR;
		case ALL_TARGETENTITIES_REMOVED:
			return ALL_ATTINSTANCES_REMOVED_STR;
		default:
			return UNKNOWN_EVENT_STR;
		}
	}

	public static int fromString(String eventTypeStr) {
		if (ELEMENT_ADDED_STR.equals(eventTypeStr))
			return ELEMENT_ADDED;
		if (ELEMENT_REMOVED_STR.equals(eventTypeStr))
			return ELEMENT_REMOVED;
		if (ATTINSTANCE_PUT_STR.equals(eventTypeStr))
			return ATTINSTANCE_PUT;
		if (ATTINSTANCE_UPDATED_STR.equals(eventTypeStr))
			return ATTINSTANCE_UPDATED;
		if (ATTINSTANCE_REMOVED_STR.equals(eventTypeStr))
			return ATTINSTANCE_REMOVED;
		if (ALL_ATTINSTANCES_REMOVED_STR.equals(eventTypeStr))
			return ALL_ATTINSTANCES_REMOVED;
		if (RELINSTANCE_PUT_STR.equals(eventTypeStr))
			return RELINSTANCE_PUT;
		if (RELINSTANCE_UPDATED_STR.equals(eventTypeStr))
			return RELINSTANCE_UPDATED;
		if (RELINSTANCE_REMOVED_STR.equals(eventTypeStr))
			return RELINSTANCE_REMOVED;
		if (ALL_RELINSTANCES_REMOVED_STR.equals(eventTypeStr))
			return ALL_RELINSTANCES_REMOVED;
		if (VALUE_UPDATED_STR.equals(eventTypeStr))
			return VALUE_UPDATED;
		if (TARGETENTITY_PUT_STR.equals(eventTypeStr))
			return TARGETENTITY_PUT;
		if (ALL_TARGETENTITIES_PUT_STR.equals(eventTypeStr))
			return ALL_TARGETENTITIES_PUT;
		if (TARGETENTITY_REMOVED_STR.equals(eventTypeStr))
			return TARGETENTITY_REMOVED;
		if (ALL_ATTINSTANCES_REMOVED_STR.equals(eventTypeStr))
			return ALL_ATTINSTANCES_REMOVED;
		return UNKNOWN_EVENT;
	}
}
