package es.ucm.fdi.grasia.faerie.core;

import java.util.List;

/**
 * Represents a relationship between a source entity and one or more target entities.
 * 
 * @param <S> the type of the source entity.
 * @param <T> the type of the target entities.
 * @author Chema
 * @see Entity
 */
public interface Relationship<S extends Entity, T extends Entity> extends ContextElement {

	/**
	 * Returns the source entity of the relationship.
	 * 
	 * @return the source entity of the relationship.
	 */
	public S getSourceEntity();

	/**
	 * Sets the source entity of the relationship.
	 * 
	 * @param sourceEntity the source entity.
	 */
	public void setSourceEntity(S sourceEntity);

	/**
	 * Returns the first target entity of the relationship. It is the same as {@code getTargetEntity(0)}.
	 * 
	 * @return the first target entity of the relationship.
	 */
	public T getTargetEntity();

	/**
	 * Returns the target entity on the specified index.
	 * 
	 * @param index the index of the entity.
	 * @return the target entity on the specified index.
	 */
	public T getTargetEntity(int index);

	/**
	 * Returns a list with all target entities.
	 * 
	 * @return a list with all target entities.
	 */
	public List<T> getAllTargetEntities();

	/**
	 * Returns the amount of target entities.
	 * 
	 * @return the amount of target entities.
	 */
	public int countTargetEntities();

	/**
	 * Sets the target entity for the index 0. It is the same as {@code setTargetEntity(0, entity)}.
	 * 
	 * @param entity the entity to set.
	 * @return the entity previously in the same index.
	 */
	public T setTargetEntity(T targetEntity);

	/**
	 * Sets the target entity for the index 0. It is the same as {@code setTargetEntity(0, entity)}.
	 * 
	 * @param index the index to put the entity.
	 * @param entity the entity to set.
	 * @return the entity previously in the same index.
	 */
	public T setTargetEntity(int index, T targetEntity);

	/**
	 * Adds a target entity at the end of the list.
	 * 
	 * @param targetEntity the target entity to be added.
	 * @return the index of the targetEntity added.
	 */
	public int addTargetEntity(T targetEntity);

	/**
	 * Resets all target entities.
	 * 
	 * @param targetEntities the new target entities.
	 */
	public void setAllTargetEntities(List<T> targetEntities);

	/**
	 * Removes the target entity on index 0. It is the same as {@code removeTargetEntity(0)}.
	 * 
	 * @return the entity previously in the same index.
	 */
	public T removeTargetEntity();

	/**
	 * Removes the target entity on the specified index.
	 * 
	 * @param index the index of the entity to be removed.
	 * @return the entity previously in the same index.
	 */
	public T removeTargetEntity(int index);

	/**
	 * Removes the specified target entity.
	 * 
	 * @param targetEntity the specified target entity.
	 * @return the index of the removed target entity.
	 */
	public int removeTargetEntity(T targetEntity);

	/**
	 * Removes all target entities.
	 */
	public void removeAllTargetEntities();
}
