package es.ucm.fdi.grasia.faerie.core.model;

import java.text.DateFormat;
import java.util.Date;

import es.ucm.fdi.grasia.faerie.core.ContextValue;
import es.ucm.fdi.grasia.faerie.core.Entity;

@SuppressWarnings("unchecked")
public class BasicContextValue<T> implements ContextValue<T> {

	private static final long serialVersionUID = 1909440711469126470L;

	private final T data;
	private final Date timeStamp = setUpTimeStamp();
	private final double quality;
	private final Entity source;

	Date setUpTimeStamp() {
		return new Date();
	}

	/**
	 * Constructs a {@code BasicContextValue} specifying the associated {@code data}, {@code quality} and {@code source}
	 * .
	 *
	 * @param data the associated data.
	 * @param quality the quality of the value, in (0, 1].
	 * @param source the entity that produced the value.
	 */
	public BasicContextValue(T data, double quality, Entity source) {
		this.data = data;
		this.quality = quality;
		this.source = source;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public T getData() {
		return data;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Date getTimeStamp() {
		return timeStamp;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public double getQuality() {
		return quality;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public <E extends Entity> E getSource() {
		return (E) source;
	}

	/**
	 * Returns the string representation of the {@code ContextValue}.
	 *
	 * @return the string representation of the {@code ContextValue}.
	 */
	@Override
	public String toString() {
		return "[" + DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT).format(timeStamp) + "] >"
				+ data + "< " + quality * 100 + "% (" + source + ")";
	}

	/**
	 * Returns the hash code of this object.
	 *
	 * @return the hash code of this object.
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 97 * hash + (this.data != null ? this.data.hashCode() : 0);
		hash = 97 * hash + (this.timeStamp != null ? this.timeStamp.hashCode() : 0);
		hash = 97 * hash
				+ (int) (Double.doubleToLongBits(this.quality) ^ (Double.doubleToLongBits(this.quality) >>> 32));
		hash = 97 * hash + (this.source != null ? this.source.hashCode() : 0);
		return hash;
	}

	/**
	 * Checks if this {@code ContextValue} represents the same value as {@code obj}.
	 *
	 * @param obj the object to check.
	 * @return true if the object represents the same {@code ContextValue}
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final BasicContextValue<?> other = (BasicContextValue<?>) obj;
		if (this.data != other.data && (this.data == null || !this.data.equals(other.data)))
			return false;
		if (this.timeStamp != other.timeStamp && (this.timeStamp == null || !this.timeStamp.equals(other.timeStamp)))
			return false;
		if (Double.doubleToLongBits(this.quality) != Double.doubleToLongBits(other.quality))
			return false;
		if (this.source != other.source)
			return false;
		return true;
	}

	@Override
	public boolean equalsWithoutDate(ContextValue obj) {
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final BasicContextValue<?> other = (BasicContextValue<?>) obj;
		if (this.data != other.data && (this.data == null || !this.data.equals(other.data)))
			return false;
		if (this.source != other.source)
			return false;
		return true;
	}
}
