package es.ucm.fdi.grasia.faerie.core.util;

import java.net.InetAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;

import es.ucm.fdi.grasia.faerie.core.Attribute;
import es.ucm.fdi.grasia.faerie.core.ContextElement;
import es.ucm.fdi.grasia.faerie.core.Relationship;

/**
 * Provides static methods to generate ids for {@code ContextElement}s.
 * 
 * @author Chema
 * @see ContextElement
 * @see URI
 */
@SuppressWarnings({ "rawtypes" })
public class ContextURIs {

	/**
	 * The scheme name for the context URIs.
	 */
	public static final String SCHEME = "context";
	/**
	 * The system property that contains the container's runnning port.
	 */
	public static final String PORT_SYSTEM_PROPERTY = "ch.ethz.iks.r_osgi.port";
	/**
	 * The default container's running port.
	 */
	public static final int PORT = 3282;
	/**
	 * Keyword to refer to any remote container.
	 */
	public static final String ANY_REMOTE_CONTAINER = "anyRemoteContainer";

	private ContextURIs() {
	}

	/**
	 * Returns the address of the local host.
	 * 
	 * @return the address of the local host.
	 */
	public static String localHostAddress() {
		try {
			InetAddress add = InetAddress.getLocalHost();
			return add.getHostAddress();
		} catch (UnknownHostException ex) {
			throw new IllegalStateException(ex);
		}
	}

	/**
	 * Returns the name of the local host.
	 * 
	 * @return the name of the local host.
	 */
	public static String localCanonicalHostName() {
		try {
			InetAddress add = InetAddress.getLocalHost();
			return add.getCanonicalHostName();
		} catch (UnknownHostException ex) {
			throw new IllegalStateException(ex);
		}
	}

	/**
	 * Returns the host name provided a host address.
	 * 
	 * @param hostAddress the host address.
	 * @return the host name.
	 * @throws UnknownHostException if the host is not found.
	 */
	public static String toCanonicalHostName(String hostAddress) throws UnknownHostException {
		InetAddress address = InetAddress.getByName(hostAddress);
		return address.getCanonicalHostName();
	}

	/**
	 * Returns the port number in which the context container is running.
	 * 
	 * @return the port number in which the context container is running.
	 */
	public static int localPort() {
		String portStr = System.getProperty(PORT_SYSTEM_PROPERTY);
		return portStr != null ? Integer.parseInt(portStr) : PORT;
	}

	public static URI localAuthority() {
		try {
			return new URI(null, null, localCanonicalHostName(), localPort(), null, null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns the context URI of the local container.
	 * 
	 * @return the context URI of the local container.
	 */
	public static URI localURI() {
		try {
			return new URI(SCHEME, null, localCanonicalHostName(), localPort(), "/", null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns the context URI representing any remote container.
	 * 
	 * @return the context URI representing any remote container.
	 */
	public static URI anyRemoteContainerURI() {
		try {
			return new URI(SCHEME, null, ANY_REMOTE_CONTAINER, -1, "/", null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Extracts the authority URI of another URI represented within a String.
	 * 
	 * @param str the string containing the URI.
	 * @return the URI containing only the authority part.
	 * @throws URISyntaxException if the specified string does not contain an URI.
	 */
	public static URI extractAuthorityURI(String str) throws URISyntaxException {
		String ssp = "//" + str.replaceAll(".*://", "").replaceAll("/.*", "");
		return new URI(null, ssp, null);
	}

	/**
	 * Returns an URI containing only the authority part of the specified URI.
	 * 
	 * @param uri the URI to extract the authority part.
	 * @return an URI containing only the authority part of the specified URI.
	 */
	public static URI authorityURI(URI uri) {
		try {
			return new URI(null, uri.getAuthority(), null, null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns an URI containing only the path part of the specified URI.
	 * 
	 * @param uri the URI to extract the path part.
	 * @return an URI containing only the path part of the specified URI.
	 */
	public static URI namespaceURI(URI uri) {
		try {
			return new URI(null, null, uri.resolve(".").getPath(), null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Returns an URI containing only the name part of the specified URI.
	 * 
	 * @param uri the URI to extract the name part.
	 * @return an URI containing only the name part of the specified URI.
	 */
	public static URI nameURI(URI uri) {
		try {
			String path = uri.getPath();
			int index = path.lastIndexOf('/');
			return new URI(null, null, path.substring(index + 1, path.length()), null, null);
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Creates an URI for an {@code Entity} with the specified parameters.
	 * 
	 * @param authority authority URI.
	 * @param path path URI.
	 * @param name name URI.
	 * @return the context URI representing the {@code Entity}.
	 */
	public static URI entityURI(URI authority, URI namespace, URI name) {
		return URI.create(SCHEME + ":/").resolve(authority).resolve(namespace).resolve(name);
	}

	/**
	 * Creates an URI for an {@code Entity} with the specified {@code String} parameters.
	 * 
	 * @param authorityStr string representing authority URI.
	 * @param namespaceStr string representing path URI.
	 * @param nameStr string representing name URI.
	 * @return the context URI representing the {@code Entity}.
	 */
	public static URI entityURI(String authorityStr, String namespaceStr, String nameStr) {
		return entityURI(URI.create(authorityStr), URI.create(namespaceStr), URI.create(nameStr));
	}

	/**
	 * Creates an URI for an {@code Entity} in the local container with the specified parameters.
	 * 
	 * @param namespace namespace URI.
	 * @param name name URI.
	 * @return the context URI representing the {@code Entity}.
	 */
	public static URI localEntityURI(URI namespace, URI name) {
		return entityURI(localAuthority(), namespace, name);
	}

	/**
	 * Creates an URI for an {@code Entity} in the local container with the specified {@code String} parameters.
	 * 
	 * @param namespaceStr string representing namespace URI.
	 * @param nameStr string representing name URI.
	 * @return the context URI representing the {@code Entity}.
	 */
	public static URI localEntityURI(String namespaceStr, String nameStr) {
		return entityURI(localAuthority().toString(), namespaceStr, nameStr);
	}

	/**
	 * Creates an URI for an {@code AttributeInstance} with the specified parameters.
	 * 
	 * @param attributeStr the string representation of the Attribute.
	 * @param entityId the URI of the entity.
	 * @return the context URI representing the {@code AttributeInstance}.
	 */
	public static URI attributeInstanceURI(String attributeStr, URI entityId) {
		return entityId.resolve("#" + attributeStr);
	}

	/**
	 * Creates an URI for an {@code AttributeInstance} with the specified parameters.
	 * 
	 * @param attributeStr the string representation of the Attribute.
	 * @param entityIdStr string representing the URI of the entity.
	 * @return the context URI representing the {@code AttributeInstance}.
	 */
	public static URI attributeInstanceURI(String attributeStr, String entityIdStr) {
		return attributeInstanceURI(attributeStr, URI.create(entityIdStr));
	}
	
	/**
	 * Creates an URI for an {@code AttributeInstance} with the specified parameters.
	 * 
	 * @param attribute the class of the Attribute.
	 * @param entityId the URI of the entity.
	 * @return the context URI representing the {@code AttributeInstance}.
	 */
	public static URI attributeInstanceURI(Class<? extends Attribute> attribute, URI entityId) {
		return attributeInstanceURI(attribute.getName(), entityId);
	}

	/**
	 * Creates an URI for an {@code AttributeInstance} with the specified parameters.
	 * 
	 * @param attribute the class of the Attribute.
	 * @param entityIdStr string representing the URI of the entity.
	 * @return the context URI representing the {@code AttributeInstance}.
	 */
	public static URI attributeInstanceURI(Class<? extends Attribute> attribute, String entityIdStr) {
		return attributeInstanceURI(attribute.getName(), URI.create(entityIdStr));
	}

	/**
	 * Creates an URI for an {@code RelationshipInstance} with the specified parameters.
	 * 
	 * @param relationshipStr the string representation of the Relationship.
	 * @param entityId the URI of the entity.
	 * @return the context URI representing the {@code RelationshipInstance}.
	 */
	public static URI relationshipInstanceURI(String relationshipStr, URI entityId) {
		return URI.create(entityId + "?" + relationshipStr);
	}

	/**
	 * Creates an URI for an {@code RelationshipInstance} with the specified parameters.
	 * 
	 * @param relationshipStr the string representation of the Relationship.
	 * @param entityIdStr string representing the URI of the entity.
	 * @return the context URI representing the {@code RelationshipInstance}.
	 */
	public static URI relationshipInstanceURI(String relationshipStr, String entityIdStr) {
		return relationshipInstanceURI(relationshipStr, URI.create(entityIdStr));
	}

	/**
	 * Creates an URI for an {@code RelationshipInstance} with the specified parameters.
	 * 
	 * @param relationship the class of the Relationship.
	 * @param entityId the URI of the entity.
	 * @return the context URI representing the {@code RelationshipInstance}.
	 */
	public static URI relationshipInstanceURI(Class<? extends Relationship> relationship, URI entityId) {
		return relationshipInstanceURI(relationship.getName(), entityId);
	}

	/**
	 * Creates an URI for an {@code RelationshipInstance} with the specified parameters.
	 * 
	 * @param relationship the class of the Relationship.
	 * @param entityIdStr string representing the URI of the entity.
	 * @return the context URI representing the {@code RelationshipInstance}.
	 */
	public static URI relationshipInstanceURI(Class<? extends Relationship> relationship, String entityIdStr) {
		return relationshipInstanceURI(relationship.getName(), URI.create(entityIdStr));
	}
}
