package es.ucm.fdi.grasia.faerie.core.util;

import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.ANY_REMOTE_CONTAINER;

import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;

import es.ucm.fdi.grasia.faerie.core.Attribute;
import es.ucm.fdi.grasia.faerie.core.Entity;
import es.ucm.fdi.grasia.faerie.core.ContextElement;
import es.ucm.fdi.grasia.faerie.core.ContextEvent;
import es.ucm.fdi.grasia.faerie.core.ContextObserver;
import es.ucm.fdi.grasia.faerie.core.Relationship;
import es.ucm.fdi.grasia.faerie.core.model.Constants;

/**
 * Provides static methods to check certain conditions of the {@code ContextEvent}s {@code URI}s.
 *
 * @author Chema
 * @see ContextObserver
 * @see ContextEvent
 */
@SuppressWarnings({ "rawtypes" })
public class Functions {

	private Functions() {
	}

	/**
	 * Checks if the specified {@code event} represents that an instance of {@code attributeStr} is added or removed
	 * from the system.
	 *
	 * @param classStr the attribute to be checked.
	 * @param event the event received.
	 * @return true if the event represents that an instance of {@code attributeStr} is added or removed, false
	 *         otherwise.
	 */
	public static boolean contextElementAddedOrRemoved(String classStr, ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			ContextElement arg = event.getContextElementArg();
			if (arg != null)
				return arg.getClass().getName().equals(classStr);
		}
		return false;
	}

	public static boolean contextElementAddedOrRemoved(Class<? extends ContextElement> contextElementClass,
			ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			ContextElement arg = event.getContextElementArg();
			return contextElementClass.isInstance(arg);
		}
		return false;
	}

	public static boolean contextElementAll(Class<? extends ContextElement> contextElementClass,
			ContextEvent event) {
		ContextElement arg = event.getContextElementArg();
		ContextElement source = event.getSource();
		return contextElementClass.isInstance(arg) || contextElementClass.isInstance(source);
	}

	/**
	 * Checks if the specified {@code event} represents that an instance of {@code attributeStr} associated with the
	 * entity with the specied {@code entityIdStr} is added or removed from the system.
	 *
	 * @param entityIdStr the id of the associated entity.
	 * @param attributeStr the attribute to be checked.
	 * @param event the event received.
	 * @return true if the event represents that an instance of {@code attributeStr} associated with the entity with the
	 *         specied {@code entityIdStr} is added or removed
	 */
	public static boolean attributeInstanceOfEntityAddedOrRemoved(String entityIdStr, String attributeStr,
			ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			Attribute arg = event.getAttributeInstanceArg();
			if (arg != null)
				return arg.getEntity().getId().toString().equals(entityIdStr)
						&& arg.getClass().getName().equals(attributeStr);
		}
		return false;
	}

	public static boolean attributeInstanceOfEntityLocalIdAddedOrRemoved(String entityLocalIdStr, String attributeStr,
			ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			Attribute arg = event.getAttributeInstanceArg();
			if (arg != null)
				return arg.getEntity().getLocalId().toString().equals(entityLocalIdStr)
						&& arg.getClass().getName().equals(attributeStr);
		}
		return false;
	}

	public static boolean relationshipInstanceOfEntityClassAddedOrRemoved(Class<? extends Entity> entityClass,
			String relationshipStr, ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			Relationship arg = event.getRelationshipInstanceArg();
			if (arg != null)
				return arg.getSourceEntity().getClass().equals(entityClass)
						&& arg.getClass().getName().equals(relationshipStr);
		}
		return false;
	}

	public static boolean relationshipInstanceOfEntityAddedOrRemoved(String entityIdStr, String relationshipStr,
			ContextEvent event) {
		if ((event.getEventType() & (ContextEvent.ELEMENT_ADDED | ContextEvent.ELEMENT_REMOVED)) != 0) {
			Relationship arg = event.getRelationshipInstanceArg();
			if (arg != null)
				return arg.getSourceEntity().getId().toString().equals(entityIdStr)
						&& arg.getClass().getName().equals(relationshipStr);
		}
		return false;
	}

	/**
	 * Returns the matching String from the two specified, or null if they does not match.
	 *
	 * @param first the first string.
	 * @param second the second string.
	 * @return the matching String from the two specified, or null if they does not match.
	 */
	public static String matchAuthority(String first, String second) {
		if (first.equals(second))
			return first;
		else if (first.equals(ANY_REMOTE_CONTAINER))
			return second;
		else if (second.equals(ANY_REMOTE_CONTAINER))
			return first;
		else
			return null;
	}

	public static Properties obtainProperties(String path) {
		try {
			Properties props = new Properties();
			FileReader reader = new FileReader(Constants.CONFIG_PATH + "/" + path);
			props.load(reader);
			reader.close();
			return props;
		} catch (IOException ex) {
			Logger.getLogger(Functions.class).error(ex.getLocalizedMessage(), ex);
			throw new RuntimeException(ex);
		}
	}

	public static String formatProperties(Map<String, Object> properties) {
		StringBuilder builder = new StringBuilder();
		for (Map.Entry<String, Object> entry : properties.entrySet()) {
			builder.append("\n\t").append(entry.getKey()).append(" = ");
			Object value = entry.getValue();
			if (value.getClass().isArray())
				builder.append(Arrays.toString((Object[]) value));
			else
				builder.append(value);
		}
		return builder.toString();
	}
}
